import random
import unittest

from textwrap import dedent

from opustrainer.modifiers.placeholders import PlaceholderTagModifier
from opustrainer.trainer import CurriculumLoader


def first(it):
  return next(iter(it))


class TestTagger(unittest.TestCase):
  def setUp(self):
    random.seed(1)

  def test_tagger_tagging(self):
    """Default mode is tagging, and will hint the target word in the source input"""
    tagger = PlaceholderTagModifier(probability=1)
    tagger.print_alignments = True
    output = tagger(['Hello world\tHallo Welt\t0-0 1-1'])
    self.assertEqual(first(output), '__source__ Hello __target__ Hallo __done__ __source__ world __target__ Welt __done__\tHallo Welt\t1-0 3-0 6-1 8-1')
    #                                ^0         ^1    ^2         ^3    ^4       ^5         ^6    ^7         ^8   ^9        ^0    ^1

  def test_tagger_replace(self):
    """Replace mode is the same as tagging mode, except that the target word
    will be random noise, teaching the model to just copy it as is."""
    tagger = PlaceholderTagModifier(probability=0.25, replace=1)
    tagger.print_alignments = True
    output = tagger(['Hello world\tHallo Welt\t0-0 1-1'])
    self.assertEqual(first(output), '__source__ Hello __target__ িৡহ __done__ world\tিৡহ Welt\t1-0 3-0 5-1')
    #                                ^0         ^1    ^2         ^3   ^4       ^5      ^0   ^1

  def test_tagger_augment(self):
    """Augment mode will add random noise without tags to both source and target
    sentence, teaching the model to copy strings it doesn't understand."""
    tagger = PlaceholderTagModifier(probability=1, augment=1)
    tagger.print_alignments = True
    output = tagger(['Hello world\tHallo Welt\t0-0 1-1'])
    self.assertEqual(first(output), 'Hello িৡহ world ЇӤӕѣѮ қӃӄЀҲ\tHallo িৡহ Welt ЇӤӕѣѮ қӃӄЀҲ\t0-0 1-1 2-2 3-3 4-4')

  def test_tagger_augment_icu(self):
    """Augment mode will add random noise without tags to both source and target
    sentence, teaching the model to copy strings it doesn't understand."""
    tagger = PlaceholderTagModifier(probability=1, augment=1, tag=0, custom_detok_src='icu:en', custom_detok_trg='icu:de')
    tagger.print_alignments = True
    output = tagger(['Hello ▁ world\tHallo ▁ Welt\t0-0 1-1 2-2'])
    self.assertEqual(first(output), 'Hello িৡহ world ټ؇ۤە٣ٮڛۃ \tHallo িৡহ Welt ټ؇ۤە٣ٮڛۃ \t0-0 1-1 2-2 3-3')


  def test_retokenize(self):
    """Pass the spm vocab to the placeholder tag generator so that it can
    retokenize the input, and update the alignments accordingly."""
    tagger = PlaceholderTagModifier(
      probability=0.25,
      custom_detok_src='en',
      custom_detok_trg='zh',
      spm_vocab_src='contrib/test-data/vocab.zhen.spm',
      spm_vocab_trg='contrib/test-data/vocab.zhen.spm') # type: ignore Path vs String type issue
    
    output = tagger(['\t'.join([
      'This is a simple test statement 🤣 .',
      #^0   ^1 ^2 ^3    ^4   ^5        ^6 ^7
      '这 是 一个 简单 的 测试 语 句 🤣 。',
      #^0 ^1 ^2  ^3   ^4 ^5   ^6 ^7 ^8 ^9
      '0-0 1-1 2-2 3-3 3-4 4-5 5-6 5-7 6-8 7-9',
    ])])
    self.assertEqual(first(output).split('\t'), [
      '__source__ This __target__ 这 __done__ is a simple test statement 🤣.',
      # [][__source__][This][ ][__target__][这][ ][__done__][ is][ a][ simple][ test][ statement][ ] []  []  []  [🤣][.]
      #^0 ^1          ^2    ^3 ^4          ^5  ^6 ^7        ^8   ^9  ^10       ^11    ^12         ^13 ^14 ^15 ^16 ^17 ^18 
      # Note the empty [] tokens before the special tokens: these are the spaces
      # that are not part of the special marker tokens. It depends on how the
      # spm vocab is trained.
      '这是一个简单的测试语句 🤣 。',
      #[这][是][一][个][简][单][的][测][试][语][句] [ ] []  []  []  [🤣][ 。]
      #^0  ^1  ^2  ^3 ^4  ^5  ^6  ^7  ^8  ^9  ^10 ^11 ^12 ^13 ^14 ^15  ^16
      '2-0 5-0 8-1 9-2 9-3 10-4 10-5 10-6 11-7 11-8 12-9 12-10 17-15 18-16',
      # 0-0 [This]      [这]    2-0
      #     [这]        [这]    5-0
      # 1-1 [is]        [是]    8-1
      # 2-2 [a]         [一个]  9-2 9-3
      # 3-3 [simple]    [简单]  10-4 10-5
      # 3-4 [simple]    [的]    10-6
      # 4-5 [test]      [测试]  11-7 11-8
      # 5-6 [statement] [语]    12-9
      # 5-7 [statement] [句]    12-10 (13-11)
      # 6-8 [🤣]        [🤣]   (14-12 15-13 16-14) 17-15
      # 7-9 [.]         [。]    18-16
    ])

  def test_augment_icu(self):
    """Pass the spm vocab to the placeholder tag generator so that it can
    retokenize the input, and update the alignments accordingly."""
    tagger = PlaceholderTagModifier(
      probability=0.2,
      augment=1,
      tag=0,
      custom_detok_src='icu:en',
      custom_detok_trg='icu:zh',
      spm_vocab_src='contrib/test-data/vocab.zhen.spm',
      spm_vocab_trg='contrib/test-data/vocab.zhen.spm')  # type: ignore Path vs String type issue

    output = tagger(['\t'.join([
      'This ▁ is ▁ a ▁ simple ▁ test ▁ statement ▁ 🤣 .',
      #^0   ^1^2 ^3^4^5^6     ^7^8   ^9^10       ^11^12^13
      '这 是 一个 简单 的 测试 语 句 ▁ 🤣 ▁ 。',
      #^0 ^1 ^2  ^3  ^4 ^5  ^6 ^7^8 ^9^10^11
      '0-0 2-1 4-2 6-3 6-4 8-5 10-6 10-7 12-9 13-11',
    ])])

    self.assertEqual(first(output).split('\t'), [
      'This িৡহ is a simple test statement 🤣.',
      # ['This', ' ', '', '', 'ি', '', '', 'ৡ', '', '', 'হ', ' is', ' a', ' simple', ' test', ' statement', ' ', '', '', '', '🤣', '.']
      '这 িৡহ 是一个简单的测试语句 🤣 。',
      # ['这', ' ', '', '', 'ি', '', '', 'ৡ', '', '', 'হ', ' 是', '一', '个', '简', '单', '的', '测', '试', '语', '句', ' ', '', '', '', '🤣', ' 。']
      '0-0 4-4 4-5 4-6 4-7 4-8 4-9 4-10 5-4 5-5 5-6 5-7 5-8 5-9 5-10 6-4 6-5 6-6 '
      '6-7 6-8 6-9 6-10 7-4 7-5 7-6 7-7 7-8 7-9 7-10 8-4 8-5 8-6 8-7 8-8 8-9 8-10 '
      '9-4 9-5 9-6 9-7 9-8 9-9 9-10 10-4 10-5 10-6 10-7 10-8 10-9 10-10 11-11 12-12 '
      '12-13 13-14 13-15 13-16 14-17 14-18 15-19 15-20 20-25 21-26'
    ])

  def test_retokenize_on_non_trigger(self):
    """Pass the spm vocab to the placeholder tag generator so that it can
    retokenize the input, even if probability is 0."""
    tagger = PlaceholderTagModifier(
      probability=0.0,
      custom_detok_src='en',
      custom_detok_trg='zh',
      spm_vocab_src='contrib/test-data/vocab.zhen.spm',
      spm_vocab_trg='contrib/test-data/vocab.zhen.spm')  # type: ignore Path vs String type issue
    
    output = tagger(['\t'.join([
      'This is a simple test statement 🤣 .',
      '这 是 一个 简单 的 测试 语 句 🤣 。',
      '0-0 1-1 2-2 3-3 3-4 4-5 5-6 5-7 6-8 7-9',
    ])])
    self.assertEqual(first(output).split('\t'), [
      'This is a simple test statement 🤣.',
      #[This][ is][ a][ simple][ test][ statement][ ] [] [] [] [🤣][.]
      #^0    ^1   ^2  ^3       ^4     ^5          ^6  ^7 ^8 ^9 ^10 ^11 
      '这是一个简单的测试语句 🤣 。',
      #[这][是][一][个][简][单][的][测][试][语][句] [ ] []  []  []  [🤣][ 。]
      #^0  ^1  ^2  ^3 ^4  ^5  ^6  ^7  ^8  ^9  ^10 ^11 ^12 ^13 ^14 ^15  ^16
      '0-0 1-1 2-2 2-3 3-4 3-5 3-6 4-7 4-8 5-9 5-10 10-15 11-16',
    ])

  def test_mode(self):
    """Test that different calls will apply different modifications when
    multiple modes are enabled."""
    tagger = PlaceholderTagModifier(
      probability=1.0,
      custom_detok_src='zh',
      augment=0.33,
      replace=0.33,
      # tag=0.33 is implicit
    )
    
    example = [
      '429 运输 中队 ( 429 野牛) , 使用 CC - 177',
      '429 Transport Squadron (429 Bison Squadron) - Flying the CC-177',
      '0-0 1-1 2-2 3-3 4-3 5-4 5-5 7-5 8-5 9-6 8-7 9-8 10-9',
    ]

    refs = [
      [ # tag + augment * 2
        '429 __source__ 运输 __target__ Transport __done__ 中队んばずがまぃぱぷろ぀゙べぢ゜そるきと (429 野牛), 使用 CC - 177 w;V|#c<X_f =L=v<ZE"Ug',
        '429 Transport Squadron んばずがまぃぱぷろ ぀゙べぢ゜そるきと (429 Bison Squadron) - Flying the CC-177 w;V|#c<X_f =L=v<ZE"Ug',
      ],
      [ # augment + tag * 2
        '429 运输 ѕӥҸӹҶҀ ӯӷѬҁӔө ҫаэшҖӹ __source__ 中队 __target__ Squadron __done__ (429 野牛), 使用 CC - __source__ 177 __target__ ビフ __done__',
        '429 Transport ѕӥҸӹҶҀ ӯӷѬҁӔө ҫаэшҖӹ Squadron (429 Bison Squadron) - Flying the ビフ',
      ],
      [ # replace * 3
        '429 __source__ 运输 __target__ ϲ΋Ιϵϔώϭ ͷϨͻξϔΛΛ __done__ __source__ 中队 __target__ ͷϨͻξϔΛΛ __done__ (429 野牛), 使用 CC - __source__ 177 __target__ ुबॹ६ॉभऺॴढ॔ ॆ्ॺढ़ऀऱ।७३ॺ __done__',
        '429 ϲ΋Ιϵϔώϭ ͷϨͻξϔΛΛ Squadron (429 Bison Squadron) - Flying ुबॹ६ॉभऺॴढ॔ ॆ्ॺढ़ऀऱ।७३ॺ CC-177',
      ]
    ]

    for ref in refs:
      output = tagger(['\t'.join(example)])
      self.assertEqual(first(output), '\t'.join(ref))

  def test_warn_if_tag_modifier_is_not_last(self):
    with self.assertLogs(level='WARNING') as logger_ctx:
      loader = CurriculumLoader()
      loader.load(dedent("""
        datasets: {}
        stages: []
        seed: 1
        modifiers:
          - Tags: 1.0
          - UpperCase: 1.0
      """))
    self.assertRegex(logger_ctx.output[0], r"Tags modifier should to be the last modifier to be applied")

  def test_exception_if_alignment_is_missing(self):
    tagger = PlaceholderTagModifier()
    with self.assertLogs(level='WARNING') as logger_ctx:
      self.assertEqual(list(tagger(['Hello world\tHallo welt\t'])), [])
      self.assertIn('IndexError', logger_ctx.output[0])

  def test_exception_if_alignment_is_invalid(self):
    tagger = PlaceholderTagModifier()
    with self.assertLogs(level='WARNING') as logger_ctx:
      self.assertEqual(list(tagger(['Hello world\tHallo welt\t0-0 1-2'])), [])
      self.assertIn('ValueError', logger_ctx.output[0])



#!/usr/bin/env python3
import heapq
import os
import subprocess
from argparse import ArgumentParser, FileType
from dataclasses import dataclass
from itertools import islice, chain
from operator import itemgetter
from queue import Queue
from random import Random
from shutil import which
from struct import Struct
from tempfile import mkstemp
from threading import Thread
from typing import TypeVar, Iterator, Iterable, List, Optional, Tuple, Callable


# Buffer size for reading files. Bufsize that Python assigns is generally too small?
BUFSIZE=2**16

# Prefer pigz if available, but fall back to calling gzip
PATH_TO_GZIP = which("pigz") or which("gzip")

HEADER = Struct('@fI') # f for random float, I for line length


@dataclass(frozen=True)
class SortTask:
	"""Job that describes to shuffle a chunk to the shuffle_chunk_worker thread.
	Passing along the seed created by the main thread because those
	random.random() calls are predictable. The order in which Shuffling tasks
	are picked up and finished may not be."""
	fileno: int
	chunk: List[Tuple[float,bytes]]

	def __call__(self) -> None:
		with os.fdopen(self.fileno, 'wb', buffering=BUFSIZE) as fh:
			self.chunk.sort(key=itemgetter(0))
			for rand, line in self.chunk:
				fh.write(HEADER.pack(rand, len(line)))
				fh.write(line)


def task_worker(queue:"Queue[Optional[Callable[[],None]]]") -> None:
	"""Worker thread that takes a queue of filenames and seeds, and shuffles them
	in memory. Put a None in the queue to make it stop."""
	while True:
		task = queue.get()
		if task is None:
			break
		task()


def iter_shuffled_file(filename:str) -> Iterable[Tuple[float,bytes]]:
	with open(filename, 'rb', buffering=BUFSIZE) as fh:
		while True:
			header = fh.read(HEADER.size)
			if header == b'':
				break
			random, length = HEADER.unpack(header)
			yield random, fh.read(length)


def shuffle(fin: Iterable[bytes], lines:int, *, seed:Optional[int]=None, threads:int=1, tmpdir:Optional[str]=None) -> Iterable[bytes]:
	"""Shuffle a list by reading it into a bunch of files (of `lines` length)
	and shuffling all of these with `threads` in-memory sorters."""
	random = Random(seed)

	chunks: List[str] = []

	try:
		if threads > 0:
			# Limiting queue to 1 pending chunk otherwise we'll run out of memory quickly.
			queue: "Queue[Optional[SortTask]]" = Queue(maxsize=threads)

			# Prepare shuffle workers to start shuffling chunks as soon as we've
			# finished writing them.
			sorters = [
				Thread(target=task_worker, args=[queue])
				for _ in range(threads)
			]

			try:
				for sorter in sorters:
					sorter.start()

				# Split the input file into separate temporary chunks
				line_it = iter(fin)
				while True:
					chunk = [(random.random(), line) for line in islice(line_it, lines)]
					if not chunk:
						break

					fileno, filename = mkstemp(dir=tmpdir)
					# Remember the chunk's filename for later
					chunks.append(filename)
					# And immediately start shuffling & writing that chunk in another thread
					# so we can use this thread to continue ingesting chunks
					queue.put(SortTask(fileno, chunk))
			finally:
				# Tell sorters that they can stop waiting
				for _ in sorters:
					queue.put(None)

				# Wait for them to finish shuffling the last files
				for sorter in sorters:
					sorter.join()
		else:
			line_it = iter(fin)
			while True:
				chunk = [(random.random(), line) for line in islice(line_it, lines)]
				if not chunk:
					break

				fileno, filename = mkstemp(dir=tmpdir)
				chunks.append(filename)

				task = SortTask(fileno, chunk)
				task()				

		# Open all chunks. We'll be reading the next line from a random one of them.
		chunk_fds = [iter_shuffled_file(filename) for filename in chunks]

		# Use heap merge to read the next smallest random element from chunk_fds
		# which are already sorted.
		for _, line in heapq.merge(*chunk_fds, key=itemgetter(0)):
			yield line

	finally:
		# Whatever happened, if a filename of a temporary file made it into the
		# `chunks` list, we are responsible for cleaning it up.
		for filename in chunks:
			os.unlink(filename)


class Reader(Iterable[bytes]):
	"""Lazily opens a file only once you start trying to read it. Also magically
	reads gzipped files."""
	def __init__(self, filename:str):
		self.filename = filename

	def _read_gzip(self, filename:str) -> Iterable[bytes]:
		"""Open gzipped files through gzip subprocess. It is faster than Python's
		gzip submodule, and you get a bit of multiprocessing for free as the
		external gzip process can decompress up to BUFSIZE while python is doing
		other things."""
		assert PATH_TO_GZIP is not None, 'No gzip executable found on system'
		child = subprocess.Popen([PATH_TO_GZIP, '-cd', filename], stdout=subprocess.PIPE, bufsize=BUFSIZE)
		assert child.stdout is not None
		yield from child.stdout
		if child.wait() != 0:
			raise RuntimeError(f'`{PATH_TO_GZIP} -cd {filename}` failed with return code {child.returncode}')

	def _read_plain(self, filename:str) -> Iterable[bytes]:
		with open(filename, 'rb') as fh:
			yield from fh

	def __iter__(self) -> Iterator[bytes]:
		if self.filename.endswith('.gz'):
			return iter(self._read_gzip(self.filename))
		else:
			return iter(self._read_plain(self.filename))


def main() -> None:
	parser = ArgumentParser()
	parser.add_argument('--batch-size', type=int, default=1_000_000, help='number of lines per chunk. Note that these chunks are read into memory when being shuffled')
	parser.add_argument('--threads', '-j', type=int, default=0, help=f'number of concurrent shuffle threads. Defaults to none')
	parser.add_argument('--temporary-directory', '-T', type=str, help='temporary directory for shuffling batches')
	parser.add_argument('--no-shuffle', '-n', action="store_false", help='Do not shuffle, to be used for debugging', dest="shuffle")
	parser.add_argument('seed', type=int)
	parser.add_argument('output', type=FileType('wb', bufsize=BUFSIZE), default='-')
	parser.add_argument('files', nargs='+')

	args = parser.parse_args()

	# Read the lines
	it: Iterable[bytes] = chain.from_iterable(Reader(filename) for filename in args.files)

	# Shuffle the lines
	if args.shuffle:
		it = shuffle(it, lines=args.batch_size, seed=args.seed, threads=args.threads, tmpdir=args.temporary_directory)

	args.output.writelines(it)


if __name__ == '__main__':
	main()
